#ifndef UTILITIEC_QUADTREE_H
#define UTILITIEC_QUADTREE_H

#include <stddef.h>

#include "../allocator-interface/allocator-interface.h"

typedef double scalar;

enum QuadTreeNodeType {
	QUADTREENODETYPE_TREE,
	QUADTREENODETYPE_LEAF,
	QUADTREENODETYPE_EMPTY,
};

typedef struct QuadSubTree_s QuadSubTree;

typedef struct QuadTreeLeaf {
	scalar x;
	scalar y;
} QuadTreeLeaf;

union QuadTreeNode {
	QuadSubTree* tree;
	QuadTreeLeaf* leaf;
};

struct QuadSubTree_s {
	enum QuadTreeNodeType top_left_type     :2;
	enum QuadTreeNodeType top_right_type    :2;
	enum QuadTreeNodeType bottom_left_type  :2;
	enum QuadTreeNodeType bottom_right_type :2;
	union QuadTreeNode top_left;
	union QuadTreeNode top_right;
	union QuadTreeNode bottom_left;
	union QuadTreeNode bottom_right;
};

typedef struct QuadTree_s {
	QuadSubTree* root;
	allocator_t* allocator;
	scalar width;
	scalar height;
} QuadTree;

int QuadTree_Create(QuadTree* tree, scalar width, scalar height, allocator_t* allocator);
int QuadTree_Insert(QuadTree* tree, QuadTreeLeaf* leaf);

typedef struct QuadTreeDimension_s {
	scalar top;
	scalar left;
	scalar right;
	scalar bottom;
} QuadTreeDimension;

typedef int (*QuadTree_TreeCallback) (void* context, QuadSubTree* tree, QuadTreeDimension dimension);
int QuadTree_ForEachTree(QuadTree* tree, QuadTree_TreeCallback action, void* context);

typedef int (*QuadTree_LeafCallback) (void* context, QuadTreeLeaf* leaf);
int QuadTree_ForEachLeaf(QuadTree* tree, QuadTree_LeafCallback action, void* context);
int QuadTree_ForEachLeafInRadius(QuadTree* tree, QuadTree_LeafCallback action, void* context, QuadTreeLeaf point, scalar radius);
void QuadTree_Destroy(QuadTree* tree);

scalar QuadTreeLeaf_Distance(QuadTreeLeaf* self, QuadTreeLeaf* other);

#endif
